from typing import Dict, Any, Optional

from fastapi_framework import disabled_modules

CONFIG_BLOCKLIST = ["CONFIG_PATH", "CONFIG_TYPE"]
CONFIG_FILE_PATH_DEFAULT = "config.yaml"
CONFIG_TYPE_DEFAULT = "yaml"


class _ConfigField:
    name: str
    type_hint: Optional[type] = None
    default_value: Any

    def __init__(self, default_value: Any = None, name: str = ""):
        self.name = name
        self.default_value = default_value


def ConfigField(default: Any = None, name: str = "") -> Any:
    return _ConfigField(default, name)


class ConfigMeta(type):
    def __new__(mcs, name, bases, dct):
        config_entries: Dict[str, _ConfigField] = {}
        config_class = super().__new__(mcs, name, bases, dct)
        annotations: Dict[str, type] = dct.get("__annotations__", {})

        for key in dct.keys():
            if key in CONFIG_BLOCKLIST:
                continue
            if not isinstance(dct[key], _ConfigField):
                continue
            type_hint: Optional[type] = None
            if key in annotations:
                type_hint = annotations[key]
            value: _ConfigField = dct[key]
            type_hint = type_hint if not getattr(type_hint, "__origin__", None) else type_hint.__origin__
            if type_hint is not None:
                try:
                    type_hint.__call__()
                except TypeError:
                    type_hint = None
            value.type_hint = type_hint
            if value.name == "":
                value.name = key
            config_entries[key] = value

        if "config" in disabled_modules:
            for config_entry in config_entries:
                setattr(config_class, config_entry, None)
            return config_class

        config_file_path = dct.get("CONFIG_PATH") or CONFIG_FILE_PATH_DEFAULT
        config_type = dct.get("CONFIG_TYPE") or CONFIG_TYPE_DEFAULT

        with open(config_file_path, "r") as file:
            data: str = file.read()

        config: Dict

        if config_type.lower() == "yaml":
            import yaml

            config = yaml.load(data, Loader=yaml.CLoader)
        elif config_type.lower() == "json":
            import json

            config = json.loads(data)
        elif config_type.lower() == "toml":
            import toml

            config = dict(toml.loads(data))
        else:
            raise Exception(f"Config Type '{config_type}' is not Supported")

        del data

        config = config or {}
        for key in config_entries.keys():
            config_key = config_entries[key].name
            if config_key in config.keys():
                type_hint = config_entries.get(key).type_hint
                if type_hint:
                    value = type_hint(config[config_key]) if config[config_key] is not None else None
                else:
                    value = config[config_key]
                setattr(config_class, key, value)
            else:
                value = config_entries.get(config_key).default_value
                setattr(config_class, key, value)

        return config_class


class Config(metaclass=ConfigMeta):
    pass



from typing import List
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, patch, MagicMock

from fastapi import HTTPException, FastAPI, Depends, Request
from httpx import AsyncClient, Response

from fastapi_framework.rate_limit import (
    RateLimitManager,
    RateLimiter,
    RateLimitTime,
    default_get_uuid,
    default_callback,
    get_uuid_user_id,
)
from fastapi_framework import rate_limit, redis_dependency

app = FastAPI()


@app.get("/limited", dependencies=[Depends(RateLimiter(2, RateLimitTime(seconds=5)))])
async def limited_route():
    return "Got it"


class TestRateLimit(IsolatedAsyncioTestCase):
    testing_uuid: str = "start_uuid"

    async def get_testing_uuid(self, _) -> str:
        return self.testing_uuid

    async def asyncSetUp(self):
        await redis_dependency.init()
        await RateLimitManager.init(await redis_dependency(), get_uuid=self.get_testing_uuid)

    @patch.object(rate_limit, "disabled_modules", [])
    async def test_rate_limit_manager_init(self):
        redis = AsyncMock()

        await RateLimitManager.init(redis)

        self.assertEqual(RateLimitManager.redis, redis)
        self.assertEqual(RateLimitManager.callback, default_callback)
        self.assertEqual(RateLimitManager.get_uuid, default_get_uuid)

    @patch.object(rate_limit, "disabled_modules", ["rate_limit"])
    async def test_rate_limit_manager_init_disabled(self):
        redis = AsyncMock()

        with self.assertRaises(Exception):
            await RateLimitManager.init(redis)

    @patch.object(rate_limit, "disabled_modules", [])
    async def test_rate_limit_time(self):
        rate_limit_time = RateLimitTime(seconds=100)

        self.assertEqual(rate_limit_time.milliseconds, 100 * 1000)

        rate_limit_time = RateLimitTime(minutes=3)

        self.assertEqual(rate_limit_time.milliseconds, 3 * 60 * 1000)

        rate_limit_time = RateLimitTime(hours=2)

        self.assertEqual(rate_limit_time.milliseconds, 2 * 60 * 60 * 1000)

        rate_limit_time = RateLimitTime(days=5)

        self.assertEqual(rate_limit_time.milliseconds, 5 * 24 * 60 * 60 * 1000)

    @patch.object(rate_limit, "disabled_modules", [])
    async def test_rate_limiter_init(self):
        count = 5
        time = RateLimitTime(seconds=56)

        rate_limiter = RateLimiter(count, time)

        self.assertEqual(rate_limiter.count, count)
        self.assertEqual(rate_limiter.time, time)
        self.assertEqual(rate_limiter.get_uuid, None)
        self.assertEqual(rate_limiter.callback, None)

    @patch.object(rate_limit, "disabled_modules", ["rate_limit"])
    async def test_rate_limiter_init_disabled(self):
        count = 5
        time = RateLimitTime(seconds=56)

        with self.assertRaises(Exception):
            RateLimiter(count, time)

    async def test_default_callback(self):
        with self.assertRaises(HTTPException):
            await default_callback({})

    async def test_default_get_uuid(self):
        host_ip = "111.222.333.444"
        request = AsyncMock()
        request.client.host = host_ip

        uuid = await default_get_uuid(request)

        self.assertIsInstance(uuid, str)
        self.assertEqual(uuid, host_ip)

    @patch("fastapi_framework.rate_limit.get_data")
    async def test_get_uuid_user_id(self, get_data_patch: AsyncMock):
        request = MagicMock()
        request.headers.get.return_value = "Bearer test_bearer_token"
        get_data_patch.return_value = {"user_id": "5"}

        uuid = await get_uuid_user_id(request)

        get_data_patch.assert_called_once_with("test_bearer_token")
        self.assertEqual(uuid, "5")

    @patch("fastapi_framework.rate_limit.get_data")
    @patch("fastapi_framework.rate_limit.HTTPBearer")
    async def test_get_uuid_user_id_no_token(self, http_bearer_patch: AsyncMock, get_data_patch: AsyncMock):
        http_bearer_patch.return_value = AsyncMock()
        http_bearer_patch.return_value.return_value = None
        request = MagicMock()

        with self.assertRaises(Exception):
            await get_uuid_user_id(request)

    async def test_limited_route(self):
        self.testing_uuid = "test_limited_route"

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/limited")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), '"Got it"')

    async def test_limited_route_without_init(self):
        self.testing_uuid = "test_limited_route_without_init"
        RateLimitManager.redis = None

        with self.assertRaises(Exception):
            async with AsyncClient(app=app, base_url="https://test") as ac:
                await ac.get("/limited")

        RateLimitManager.redis = await redis_dependency()

    async def test_spam_limited_route_with_async_callback(self):
        self.testing_uuid = "test_spam_limited_route_with_async_callback"
        async_callback = AsyncMock()
        RateLimitManager.callback = async_callback

        async with AsyncClient(app=app, base_url="https://test") as ac:
            for i in range(4):
                response: Response = await ac.get("/limited")

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode("utf-8"), '"Got it"')
            if i >= 2:
                async_callback.assert_called()

        RateLimitManager.callback = default_callback

    async def test_spam_limited_route(self):
        self.testing_uuid = "test_spam_limited_route"
        responses: List[Response] = []

        async with AsyncClient(app=app, base_url="https://test") as ac:
            for _ in range(4):
                responses.append(await ac.get("/limited"))

        for i, response in enumerate(responses):
            if i < 2:
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content.decode("utf-8"), '"Got it"')
            else:
                self.assertEqual(response.status_code, 429)
                self.assertTrue("detail" in response.json())
                self.assertEqual(response.json()["detail"], "Too Many Requests")



from typing import Union, Callable, Dict, Coroutine, Optional, Any

from fastapi import Request, HTTPException, Response
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials

from .in_memory_backend import InMemoryBackend
from .jwt_auth import get_data
from .modules import disabled_modules


async def default_callback(headers: Dict):
    """Default Error Callback when get Raid Limited"""
    raise HTTPException(429, detail="Too Many Requests", headers=headers)


async def default_get_uuid(request: Request) -> str:
    """Default getter for UUID working with Users IP"""
    return f"{request.client.host}"


async def get_uuid_user_id(request: Request):
    """Getter for UUID working with User IDs from the JWTs"""
    bearer_auth: Optional[HTTPAuthorizationCredentials] = await (HTTPBearer())(request)
    if not bearer_auth:
        raise Exception("Cant get HTTPBearer Auth Token")
    token: str = bearer_auth.credentials
    data: Dict = await get_data(token)
    return f"{data['user_id']}"


class RateLimitManager:
    """Rate Limit Manager for Redis, UUID Getter and the Error Callback"""

    redis: InMemoryBackend
    get_uuid: Callable = default_get_uuid
    callback: Callable = default_callback

    @classmethod
    async def init(
        cls,
        redis: InMemoryBackend,
        get_uuid: Callable = default_get_uuid,
        callback: Callable = default_callback,
    ):
        """Initialise Rate Limit Manager"""
        if "rate_limit" in disabled_modules:
            raise Exception("Module Rate Limit is disabled")
        cls.redis = redis
        cls.get_uuid = get_uuid
        cls.callback = callback


class RateLimitTime:
    __milliseconds: int

    def __init__(self, seconds: int = 0, minutes: int = 0, hours: int = 0, days: int = 0):
        self.__milliseconds = (seconds + (minutes * 60) + (hours * 60 * 60) + (days * 60 * 60 * 24)) * 1000

    @property
    def milliseconds(self):
        return self.__milliseconds


class RateLimiter:
    """Raid Limit Dependency"""

    count: int
    time: RateLimitTime
    get_uuid: Union[Callable, None]
    callback: Union[Callable, None]

    def __init__(
        self,
        count,
        time: RateLimitTime,
        get_uuid: Union[Callable, None] = None,
        callback: Union[Callable, None] = None,
    ):
        if "rate_limit" in disabled_modules:
            raise Exception("Module Rate Limit is disabled")
        self.time = time
        self.count = count
        self.get_uuid = get_uuid
        self.callback = callback

    async def __call__(self, request: Request, response: Response):
        if not RateLimitManager.redis:
            raise Exception("You have to initialise the RateLimitManager at the Startup")

        get_uuid: Union[Callable] = self.get_uuid or RateLimitManager.get_uuid
        callback: Union[Callable] = self.callback or RateLimitManager.callback
        uuid: Union[str, Coroutine] = get_uuid(request)

        if isinstance(uuid, Coroutine):
            uuid = await uuid
        redis_key: str = f"rate_limit:{request.url.path}:{uuid}"
        redis_key_lock: str = f"{redis_key}:lock"
        if await RateLimitManager.redis.exists(redis_key_lock):
            headers = await self.get_headers(redis_key)
            result: Any = callback(headers)
            if isinstance(result, Coroutine):
                await result
            return

        count = await RateLimitManager.redis.incr(redis_key)
        if count == 1:
            await RateLimitManager.redis.pexpire(redis_key, self.time.milliseconds)
        if count >= self.count:
            pttl: int = await RateLimitManager.redis.pttl(redis_key)
            await RateLimitManager.redis.delete(redis_key)
            await RateLimitManager.redis.set(redis_key_lock, 1, pexpire=pttl)
        headers = await self.get_headers(redis_key)
        for key in headers.keys():
            response.headers[key] = headers[key]

    async def get_headers(self, redis_key: str) -> Dict:
        """Generates Rate Limit Headers"""
        headers: Dict = {}
        redis_value = await RateLimitManager.redis.get(redis_key)
        redis_value = redis_value if redis_value is None else redis_value.decode("utf-8")
        headers["X-Rate-Limit-Limit"] = f"{self.count}"
        headers["X-Rate-Limit-Remaining"] = str(
            redis_value or (0 if await RateLimitManager.redis.exists(f"{redis_key}:lock") else self.count)
        )
        ttl: int = await RateLimitManager.redis.ttl(redis_key)
        ttl = ttl if ttl != -2 else await RateLimitManager.redis.ttl(f"{redis_key}:lock")
        ttl = ttl if ttl != -2 else 0
        headers["X-Rate-Limit-Reset"] = f"{ttl}"
        return headers



from datetime import timedelta, datetime
from typing import Dict, Union, List
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock, patch, AsyncMock

import jwt
from aioredis import Redis
from fastapi import FastAPI, Depends, HTTPException
from httpx import AsyncClient, Response

from fastapi_framework import redis_dependency
from fastapi_framework.jwt_auth import (
    get_token,
    create_jwt_token,
    ALGORITHM,
    get_data,
    create_access_token,
    create_refresh_token,
    invalidate_refresh_token,
    check_refresh_token,
    generate_tokens,
)

app = FastAPI()

users: List[Dict[str, Union[str, int]]] = [
    {"user_id": 1, "username": "test", "password": "123"},
    {"user_id": 2, "username": "admin", "password": "AdminPassword"},
]


@app.get("/token")
async def token_route(username: str, password: str, redis: Redis = Depends(redis_dependency)):
    user: Union[Dict[str, Union[str, int]], None] = None
    for user_item in users:
        if user_item["username"] == username:
            user = user_item.copy()  # noqa
            break

    if (user is None) or password != user["password"]:
        raise HTTPException(401, detail="Username or Password is wrong")
    return await generate_tokens(
        {"user": {"id": user["user_id"], "username": user["username"]}}, int(user["user_id"]), redis
    )


@app.get("/refresh")
async def refresh_route(refresh_token: str, redis: Redis = Depends(redis_dependency)):
    data: Dict = {}
    if not await check_refresh_token(refresh_token, redis):
        raise HTTPException(401, "Refresh Token Invalid")
    try:
        data = await get_data(refresh_token)
    except HTTPException as e:
        if e.detail == "Token is expired":
            await invalidate_refresh_token(refresh_token, redis)
            raise e
    user_id = int(data["user_id"])
    username: str = ""
    for user in users:
        if user["user_id"] == user_id:
            username = str(user["username"])
    await invalidate_refresh_token(refresh_token, redis)
    return await generate_tokens({"user": {"id": user_id, "username": username}}, int(user_id), redis)


@app.get("/logout", dependencies=[Depends(get_token)])
async def logout_route(refresh_token: str, redis: Redis = Depends(redis_dependency)):
    await invalidate_refresh_token(refresh_token, redis)


@app.get("/secret", dependencies=[Depends(get_data)])
async def secured_route():
    return "Hello!"


class TestJWTAuth(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        await redis_dependency.init()

    async def test_get_token(self):
        bearer_scheme = MagicMock()
        bearer_scheme.credentials = "TEST_JWT_TOKEN"
        self.assertEqual(await get_token(bearer_scheme), "TEST_JWT_TOKEN")

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_create_jwt_token(self):
        data = {"test": "test_value"}

        jwt_token = await create_jwt_token(data, timedelta(minutes=30))

        self.assertIsInstance(jwt_token, str)
        decoded_data = jwt.decode(jwt_token, "TEST_SECRET_KEY", algorithms=[ALGORITHM])
        self.assertTrue("test" in decoded_data)
        self.assertTrue("exp" in decoded_data)
        self.assertEqual(decoded_data["test"], data["test"])
        self.assertIsInstance(decoded_data["exp"], int)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_get_data(self):
        data = {"test": "test_value"}

        jwt_token = jwt.encode(data, "TEST_SECRET_KEY", algorithm=ALGORITHM)

        decoded_data = await get_data(jwt_token)
        self.assertIsInstance(decoded_data, Dict)
        self.assertEqual(decoded_data, data)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_get_data_invalid_token(self):
        data = {"exp": datetime.utcnow() - timedelta(minutes=10)}

        jwt_token = jwt.encode(data, "TEST_SECRET_KEY", algorithm=ALGORITHM)

        with self.assertRaises(HTTPException):
            await get_data(jwt_token)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_create_access_token(self):
        data = {"test": "test_value"}

        jwt_token = await create_access_token(data)

        decoded_data = jwt.decode(jwt_token, "TEST_SECRET_KEY", algorithms=[ALGORITHM])
        self.assertTrue("test" in decoded_data)
        self.assertEqual(decoded_data["test"], data["test"])

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_create_refresh_token(self):
        user_id: int = 5
        redis = AsyncMock()

        jwt_token = await create_refresh_token(user_id, redis)

        redis.sadd.assert_called_once_with("refresh_tokens", jwt_token)
        decoded_data = jwt.decode(jwt_token, "TEST_SECRET_KEY", algorithms=[ALGORITHM])
        self.assertTrue("user_id" in decoded_data)
        self.assertEqual(decoded_data["user_id"], user_id)

    async def test_invalidate_refresh_token(self):
        redis = AsyncMock()

        await invalidate_refresh_token("TEST_REFRESH_TOKEN", redis)

        redis.srem.assert_called_once_with("refresh_tokens", "TEST_REFRESH_TOKEN")

    async def test_check_refresh_token_positive(self):
        redis = AsyncMock()
        redis.smembers = AsyncMock()
        redis.smembers.return_value = [
            b"TEST_FALSE_REFRESH_TOKEN",
            b"TEST_SECOND_FALSE_REFRESH_TOKEN",
            b"TEST_REFRESH_TOKEN",
        ]

        result = await check_refresh_token("TEST_REFRESH_TOKEN", redis)

        redis.smembers.assert_called_once_with("refresh_tokens")
        self.assertTrue(result)

    async def test_check_refresh_token_negative(self):
        redis = AsyncMock()
        redis.smembers = AsyncMock()
        redis.smembers.return_value = [
            b"TEST_FALSE_REFRESH_TOKEN",
            b"TEST_SECOND_FALSE_REFRESH_TOKEN",
        ]

        result = await check_refresh_token("TEST_REFRESH_TOKEN", redis)

        redis.smembers.assert_called_once_with("refresh_tokens")
        self.assertFalse(result)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_generate_tokens(self):
        data = {"test": "test_value"}
        user_id = 6
        redis = AsyncMock()

        tokens = await generate_tokens(data, user_id, redis)

        self.assertTrue("token_type" in tokens)
        self.assertEqual(tokens["token_type"], "bearer")
        self.assertTrue("access_token" in tokens)
        self.assertTrue("refresh_token" in tokens)
        access_token = tokens["access_token"]
        refresh_token = tokens["refresh_token"]
        redis.sadd.assert_called_once_with("refresh_tokens", refresh_token)
        decoded_access_token = jwt.decode(access_token, "TEST_SECRET_KEY", algorithms=[ALGORITHM])
        self.assertTrue("test" in decoded_access_token)
        self.assertEqual(decoded_access_token["test"], data["test"])
        decoded_refresh_token = jwt.decode(refresh_token, "TEST_SECRET_KEY", algorithms=[ALGORITHM])
        self.assertTrue("user_id" in decoded_refresh_token)
        self.assertEqual(decoded_refresh_token["user_id"], user_id)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_login(self):
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/token", params={"username": "test", "password": "123"})

        self.assertEqual(response.status_code, 200)
        self.assertTrue("access_token" in response.json())
        self.assertTrue("refresh_token" in response.json())
        self.assertTrue("token_type" in response.json())
        self.assertEqual(response.json()["token_type"], "bearer")

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_login_invalid_credentials(self):
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/token", params={"username": "not_exists", "password": "wrong"})

        self.assertEqual(response.status_code, 401)

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "TEST_SECRET_KEY")
    async def test_secret_route(self):
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/token", params={"username": "test", "password": "123"})

        self.assertEqual(response.status_code, 200)
        self.assertTrue("access_token" in response.json())
        self.assertTrue("token_type" in response.json())
        self.assertEqual(response.json()["token_type"], "bearer")
        access_token = response.json()["access_token"]

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/secret", headers={"Authorization": f"Bearer {access_token}"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), '"Hello!"')

    @patch("fastapi_framework.jwt_auth.SECRET_KEY", "")
    async def test_empty_secret_key(self):
        from fastapi_framework.jwt_auth import check_secret_key

        with self.assertRaises(Exception):
            check_secret_key()



import time
from typing import Dict, Any, Optional, Set, Union, List
from abc import ABC


class InMemoryBackend(ABC):
    async def set(self, key: str, value, expire: int = 0, pexpire: int = 0, exists=None):
        """Set Key to Value"""

    async def get(self, key: str):
        """Get Value from Key"""

    async def pttl(self, key: str) -> int:
        """Get PTTL from a Key"""

    async def ttl(self, key: str) -> int:
        """Get TTL from a Key"""

    async def pexpire(self, key: str, pexpire: int) -> bool:
        """Sets and PTTL for a Key"""

    async def expire(self, key: str, expire: int) -> bool:
        """Sets and TTL for a Key"""

    async def incr(self, key: str) -> int:
        """Increases an Int Key"""

    async def decr(self, key: str) -> int:
        """Decreases an Int Key"""

    async def delete(self, key: str):
        """Delete value of a Key"""

    async def smembers(self, key: str) -> Set:
        """Gets Set Members"""

    async def sadd(self, key: str, value: Any) -> bool:
        """Adds a Member to a Set"""

    async def srem(self, key: str, member: Any) -> bool:
        """Removes a Member from a Set"""

    async def exists(self, key: str) -> bool:
        """Checks if a Key exists"""


class RAMBackendItem:
    """Key-Value Item for the RAM Backend"""

    value: Union[bytes, List]
    pexpire: int
    timestamp: int

    def __init__(self, value: Any, pexpire: int):
        self.value = value
        self.pexpire = pexpire
        self.timestamp = int(time.time() * 1000)


class RAMBackend(InMemoryBackend):
    """Python In Memory Backend"""

    data: Dict[str, RAMBackendItem] = {}
    SET_IF_NOT_EXIST = "SET_IF_NOT_EXIST"  # NX
    SET_IF_EXIST = "SET_IF_EXIST"  # XX

    async def _check_key_expire(self, key: str, item: RAMBackendItem):
        """Checks if a Key is expired and deletes it"""
        timestamp = int(time.time() * 1000)
        if item.pexpire > 0 and (item.timestamp + item.pexpire) <= timestamp:
            self.data.pop(key)
            return False
        return True

    async def set(self, key: str, value: Any, expire: int = 0, pexpire: int = 0, exists=None):
        """Set Key to Value"""
        if not isinstance(value, bytes) and not isinstance(value, List):
            value = bytes(str(value), "utf-8")
        if exists == self.SET_IF_NOT_EXIST:
            if key in self.data:
                return
        elif exists == self.SET_IF_EXIST:
            if key not in self.data:
                return
        elif exists is None:
            pass
        else:
            raise Exception("Wrong Params")
        self.data[key] = RAMBackendItem(value, pexpire + (expire * 1000))

    async def get(self, key: str):
        """Get Value from Key"""
        item: Optional[RAMBackendItem] = self.data.get(key)
        if not item:
            return None
        if not await self._check_key_expire(key, item):
            return None
        return item.value

    async def pttl(self, key: str) -> int:
        """Get PTTL from a Key"""
        item: Optional[RAMBackendItem] = self.data.get(key)
        timestamp = int(time.time() * 1000)
        if not item:
            return -2
        if not await self._check_key_expire(key, item):
            return -1
        if item.pexpire == 0:
            return -1
        return (item.pexpire + item.timestamp) - timestamp

    async def ttl(self, key: str) -> int:
        """Get TTL from a Key"""
        pttl = await self.pttl(key)
        if pttl >= 0:
            return pttl // 1000
        return pttl

    async def pexpire(self, key: str, pexpire: int) -> bool:
        """Sets and PTTL for a Key"""
        item: Optional[RAMBackendItem] = self.data.get(key)
        timestamp = int(time.time() * 1000)
        if not item:
            return False
        if not await self._check_key_expire(key, item):
            return False
        item.timestamp = timestamp
        item.pexpire = pexpire
        self.data[key] = item
        return True

    async def expire(self, key: str, expire: int) -> bool:
        """Sets and TTL for a Key"""
        return await self.pexpire(key, expire * 1000)

    async def incr(self, key: str) -> int:
        """Increases an Int Key"""
        item: Optional[RAMBackendItem] = self.data.get(key)
        if not item:
            await self.set(key, 1)
            return 1
        try:
            if not isinstance(item.value, bytes):
                raise Exception("Value must be a Int")
            item.value = bytes(str(int(item.value.decode("utf-8")) + 1), "utf-8")
        except ValueError:
            raise Exception("Value must be a Int")
        self.data[key] = item
        return int(item.value.decode("utf-8"))

    async def decr(self, key: str) -> int:
        """Decreases an Int Key"""
        item: Optional[RAMBackendItem] = self.data.get(key)
        if not item:
            await self.set(key, -1)
            return -1
        try:
            if not isinstance(item.value, bytes):
                raise Exception("Value must be a Int")
            item.value = bytes(str(int(item.value.decode("utf-8")) - 1), "utf-8")
        except ValueError:
            raise Exception("Value must be a Int")
        self.data[key] = item
        return int(item.value.decode("utf-8"))

    async def delete(self, key: str):
        """Delete value of a Key"""
        if key not in self.data:
            return
        del self.data[key]

    async def smembers(self, key: str) -> Set:
        """Gets Set Members"""
        data: Optional[Union[bytes, List]] = await self.get(key)
        if not data:
            return set()
        if not isinstance(data, list):
            return {data}
        return set(data)

    async def sadd(self, key: str, value: Any) -> bool:
        """Adds a Member to a Set"""
        data: Union[Optional[Union[bytes, List]], Set] = await self.get(key)
        if not data or not isinstance(data, List):
            data = {value}
        else:
            data = set(data)
            data.add(value)
        await self.set(key, list(data))
        return True

    async def srem(self, key: str, member: Any) -> bool:
        """Removes a Member from a Set"""
        data: Union[Optional[Union[bytes, List]], Set] = await self.get(key)
        if not data or not isinstance(data, List):
            return False
        data = set(data)
        if member not in data:
            return False
        data.remove(member)
        await self.set(key, list(data))
        return True

    async def exists(self, key: str) -> bool:
        """Checks if a Key exists"""
        return key in self.data



from typing import Set, Any, Optional

from aioredis import create_redis_pool
from aioredis import Redis as RedisConnection
from dotenv import load_dotenv
from os import getenv
from .in_memory_backend import InMemoryBackend, RAMBackend

from .modules import disabled_modules

load_dotenv()

Redis = InMemoryBackend

REDIS_HOST = getenv("REDIS_HOST", "localhost")
REDIS_PORT = getenv("REDIS_PORT", "6379")


class RedisBackend(InMemoryBackend):
    redis_connection: RedisConnection

    @staticmethod
    async def init(url: str) -> "RedisBackend":
        redis = RedisBackend()
        redis.redis_connection = await create_redis_pool(url)
        return redis

    async def get(self, key: str):
        """Get Value from Key"""
        return await self.redis_connection.get(key)

    async def set(self, key: str, value, expire: int = 0, pexpire: int = 0, exists=None):
        """Set Key to Value"""
        return await self.redis_connection.set(key, value, expire=expire, pexpire=pexpire, exist=exists)

    async def pttl(self, key: str) -> int:
        """Get PTTL from a Key"""
        return int(await self.redis_connection.pttl(key))

    async def ttl(self, key: str) -> int:
        """Get TTL from a Key"""
        return int(await self.redis_connection.ttl(key))

    async def pexpire(self, key: str, pexpire: int) -> bool:
        """Sets and PTTL for a Key"""
        return bool(await self.redis_connection.pexpire(key, pexpire))

    async def expire(self, key: str, expire: int) -> bool:
        """Sets and TTL for a Key"""
        return bool(await self.redis_connection.expire(key, expire))

    async def incr(self, key: str) -> int:
        """Increases an Int Key"""
        return int(await self.redis_connection.incr(key))

    async def decr(self, key: str) -> int:
        """Decreases an Int Key"""
        return int(await self.redis_connection.decr(key))

    async def delete(self, key: str):
        """Delete value of a Key"""
        return await self.redis_connection.delete(key)

    async def smembers(self, key: str) -> Set:
        """Gets Set Members"""
        return set(await self.redis_connection.smembers(key))

    async def sadd(self, key: str, value: Any) -> bool:
        """Adds a Member to a Dict"""
        return bool(await self.redis_connection.sadd(key, value))

    async def srem(self, key: str, member: Any) -> bool:
        """Removes a Member from a Set"""
        return bool(await self.redis_connection.srem(key, member))

    async def exists(self, key: str) -> bool:
        """Checks if a Key exists"""
        return bool(await self.redis_connection.exists(key))


class RedisDependency:
    """FastAPI Dependency for Redis Connections"""

    redis: Optional[InMemoryBackend] = None

    async def __call__(self):
        return self.redis

    async def init(self):
        """Initialises the Redis Dependency"""
        if "redis" in disabled_modules:
            self.redis = RAMBackend()
        else:
            self.redis = await RedisBackend.init(f"redis://{REDIS_HOST}:{REDIS_PORT}")


redis_dependency: RedisDependency = RedisDependency()


async def get_redis() -> RedisBackend:
    """Returns a NEW Redis connection"""
    if "redis" in disabled_modules:
        raise Exception("Module Redis is disabled")
    return await RedisBackend.init(f"redis://{REDIS_HOST}:{REDIS_PORT}")



from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch, MagicMock, AsyncMock

from fastapi_framework.redis import RedisDependency, get_redis, RedisBackend
from fastapi_framework import redis, RAMBackend


class TestRedis(IsolatedAsyncioTestCase):
    @patch.object(redis, "disabled_modules", [])
    async def test_redis_dependency_init(self):
        redis_dependency: RedisDependency = RedisDependency()

        self.assertEqual(redis_dependency.redis, None)

        await redis_dependency.init()

        self.assertNotEqual(redis_dependency.redis, None)
        self.assertIsInstance(redis_dependency.redis, RedisBackend)

    @patch.object(redis, "disabled_modules", ["redis"])
    async def test_redis_dependency_init_disabled_redis(self):
        redis_dependency: RedisDependency = RedisDependency()

        await redis_dependency.init()

        self.assertIsInstance(redis_dependency.redis, RAMBackend)

    async def test_redis_dependency_call(self):
        redis_dependency: RedisDependency = RedisDependency()
        redis_dependency.redis = MagicMock()

        self.assertEqual(await redis_dependency.__call__(), redis_dependency.redis)

    @patch.object(redis, "disabled_modules", [])
    async def test_get_redis(self):
        result = await get_redis()

        self.assertIsInstance(result, RedisBackend)

    @patch.object(redis, "disabled_modules", ["redis"])
    async def test_get_redis_disabled_redis(self):
        with self.assertRaises(Exception):
            await get_redis()

    @patch.object(redis, "disabled_modules", [])
    async def test_expire(self):
        redis_backend = RedisBackend()
        redis_backend.redis_connection = AsyncMock()

        await redis_backend.expire("test", 5000)

        redis_backend.redis_connection.expire.assert_called_with("test", 5000)

    @patch.object(redis, "disabled_modules", [])
    async def test_decrease(self):
        redis_backend = RedisBackend()
        redis_backend.redis_connection = AsyncMock()

        await redis_backend.decr("test")

        redis_backend.redis_connection.decr.assert_called_with("test")

    @patch.object(redis, "disabled_modules", [])
    async def test_smembers(self):
        redis_backend = RedisBackend()
        redis_backend.redis_connection = AsyncMock()

        await redis_backend.smembers("test")

        redis_backend.redis_connection.smembers.assert_called_with("test")

    @patch.object(redis, "disabled_modules", [])
    async def test_sadd(self):
        redis_backend = RedisBackend()
        redis_backend.redis_connection = AsyncMock()

        await redis_backend.sadd("test", "test_value")

        redis_backend.redis_connection.sadd.assert_called_with("test", "test_value")

    @patch.object(redis, "disabled_modules", [])
    async def test_srem(self):
        redis_backend = RedisBackend()
        redis_backend.redis_connection = AsyncMock()

        await redis_backend.srem("test", "test_value")

        redis_backend.redis_connection.srem.assert_called_with("test", "test_value")



from typing import Any
from unittest import TestCase
from unittest.mock import patch, mock_open
from fastapi_framework import Config, ConfigField, config


class TestConfig(TestCase):
    test_yaml_config_data = "some_data: a value"
    test_json_config_data = '{"some_data": "a value"}'
    test_toml_config_data = 'some_data = "a value"'
    test_yaml_config_with_typing_data = "items:\n  - an item\n  - an other item"

    @patch("builtins.open", mock_open(read_data=test_yaml_config_data))
    def test_yaml_config(self):
        class MyConfig(Config):
            some_data: str = ConfigField("some_data")
            some_other_data: str = ConfigField(default="default value")

        self.assertEqual(MyConfig.some_data, "a value")
        self.assertEqual(MyConfig.some_other_data, "default value")

    @patch("builtins.open", mock_open(read_data=test_json_config_data))
    def test_json_config(self):
        class MyConfig(Config):
            CONFIG_PATH = "config.json"
            CONFIG_TYPE = "json"

            some_data: str = ConfigField("some_data")
            some_other_data: str = ConfigField(default="default value")

        self.assertEqual(MyConfig.some_data, "a value")
        self.assertEqual(MyConfig.some_other_data, "default value")

    @patch("builtins.open", mock_open(read_data=test_toml_config_data))
    def test_toml_config(self):
        class MyConfig(Config):
            CONFIG_PATH = "config.toml"
            CONFIG_TYPE = "toml"

            some_data: str = ConfigField("some_data")
            some_other_data: str = ConfigField(default="default value")

        self.assertEqual(MyConfig.some_data, "a value")
        self.assertEqual(MyConfig.some_other_data, "default value")

    @patch("builtins.open", mock_open(read_data=test_yaml_config_data))
    def test_yaml_config_with_blocked_name(self):
        class MyConfig(Config):
            CONFIG_PATH: str = "config.yaml"
            CONFIG_TYPE: str = "yaml"

            some_data: str = ConfigField("some_data")
            some_other_data: str = ConfigField(default="default value")

        self.assertEqual(MyConfig.some_data, "a value")
        self.assertEqual(MyConfig.some_other_data, "default value")

    @patch("builtins.open", mock_open(read_data=test_yaml_config_data))
    def test_config_type_dont_exists(self):
        with self.assertRaises(Exception):

            class _(Config):
                CONFIG_TYPE = "type doesn't exists"

                some_data: str = ConfigField()
                some_other_data: str = ConfigField(default="default value")

    @patch("builtins.open", mock_open(read_data=test_yaml_config_with_typing_data))
    def test_yaml_config_with_typing(self):
        from typing import List

        class MyConfig(Config):
            items: List = ConfigField()

        self.assertEqual(MyConfig.items, ["an item", "an other item"])

    @patch("builtins.open", mock_open(read_data=test_yaml_config_data))
    def test_yaml_config_with_type_any(self):
        class _(Config):
            some_data: Any = ConfigField("some_data")
            some_other_data: Any = ConfigField(default="default value")

    @patch.object(config, "disabled_modules", ["config"])
    def test_config_disabled(self):
        class MyConfig(Config):
            some_data: str = ConfigField("some_data")
            some_other_data: str = ConfigField(default="default value")

        self.assertEqual(MyConfig.some_data, None)
        self.assertEqual(MyConfig.some_other_data, None)



from unittest import IsolatedAsyncioTestCase

from fastapi_framework import redis_dependency, Redis
from fastapi_framework.settings import SettingsModel, Settings
from fastapi_framework.database import database_dependency, DB, select


class TestSettings(IsolatedAsyncioTestCase):
    async def test_settings_set(self):
        db: DB = await database_dependency()
        redis: Redis = await redis_dependency()

        await Settings.set("test_settings_set", "test")

        self.assertEqual((await db.first(select(SettingsModel).filter_by(key="test_settings_set"))).value, "test")
        self.assertEqual(await redis.get("settings:test_settings_set"), b"test")

    async def test_settings_get(self):
        await Settings.set("test_settings_get", "test_value")

        result = await Settings.get("test_settings_get")

        self.assertEqual(result, "test_value")

    async def test_settings_set_with_bool(self):
        db: DB = await database_dependency()
        redis: Redis = await redis_dependency()

        await Settings.set("test_settings_set_with_bool", True)
        await Settings.set("test_settings_set_with_bool_false", False)

        self.assertEqual(
            (await db.first(select(SettingsModel).filter_by(key="test_settings_set_with_bool"))).value, "1"
        )
        self.assertEqual(await redis.get("settings:test_settings_set_with_bool"), b"1")
        self.assertEqual(
            (await db.first(select(SettingsModel).filter_by(key="test_settings_set_with_bool_false"))).value, "0"
        )
        self.assertEqual(await redis.get("settings:test_settings_set_with_bool_false"), b"0")

    async def test_settings_change_value(self):
        await Settings.set("test_settings_change_value", "old_value")
        await Settings.set("test_settings_change_value", "test_value")

        self.assertEqual(await Settings.get("test_settings_change_value"), "test_value")

    async def test_settings_get_uncached_value(self):
        redis: Redis = await redis_dependency()

        await Settings.set("test_settings_get_uncached_value", "test_value")
        await redis.delete("settings:test_settings_get_uncached_value")
        result = await Settings.get("test_settings_get_uncached_value")

        self.assertEqual(result, "test_value")

    async def test_settings_get_not_exists(self):
        result = await Settings.get("test_settings_get_not_exists")

        self.assertEqual(result, None)



from os import getenv
from typing import Union, List, Dict
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock, patch

from fastapi import HTTPException, FastAPI, Depends
from sqlalchemy import Column, String, Integer

from fastapi_framework.database import (
    select,
    filter_by,
    exists,
    delete,
    database_dependency,
    DB,
    DatabaseDependency,
    Base,
)

from httpx import AsyncClient, Response
from random import choices
from string import ascii_letters

app = FastAPI()


class User(Base):
    __tablename__ = "users"
    id: Union[Column, int] = Column(Integer, primary_key=True)
    name: Union[Column, str] = Column(String(255))

    @staticmethod
    async def create(name: str) -> "User":
        db: DB = await database_dependency()
        row = User(name=name)
        await db.add(row)
        return row


@app.get("/users")
async def get_users(db: DB = Depends(database_dependency)):
    return await db.all(select(User))


@app.get("/users/{name}")
async def get_user_by_name(name: str, db: DB = Depends(database_dependency)):
    return await db.all(select(User).filter_by(name=name))


@app.post("/users/{name}")
async def add_user(name: str, db: DB = Depends(database_dependency)) -> User:
    if await db.exists(select(User).filter_by(name=name)):
        raise HTTPException(409, "Username already used")
    user = await User.create(name)
    await db.commit()
    return user


@app.delete("/users/{name}")
async def remove_user(name: str, db: DB = Depends(database_dependency)):
    query = select(User).filter_by(name=name)
    if not await db.exists(query):
        raise HTTPException(404, "User doesn't exists")
    await db.delete(await db.first(query))
    return True


class TestDatabase(IsolatedAsyncioTestCase):
    @patch("fastapi_framework.database.sa_select")
    async def test_select(self, sa_select: MagicMock):
        model = MagicMock()

        result = select(model)

        sa_select.assert_called_with(model)
        self.assertEqual(result, sa_select())

    @patch("fastapi_framework.database.select")
    async def test_filter_by(self, select_patch: MagicMock):
        model = MagicMock()
        filters = {
            "param1": MagicMock(),
            "param2": MagicMock(),
            "param3": MagicMock(),
            "param4": MagicMock(),
            "param5": MagicMock(),
        }

        result = filter_by(model, **filters)

        select_patch.assert_called_once_with(model)
        select_patch().filter_by.assert_called_once_with(**filters)
        self.assertEqual(select_patch().filter_by(), result)

    @patch("fastapi_framework.database.sa_exists")
    async def test_exists(self, sa_exists: MagicMock):
        model = MagicMock()
        filters = {
            "param1": MagicMock(),
            "param2": MagicMock(),
            "param3": MagicMock(),
            "param4": MagicMock(),
            "param5": MagicMock(),
        }

        result = exists(model, **filters)

        sa_exists.assert_called_once_with(model, **filters)
        self.assertEqual(result, sa_exists())

    @patch("fastapi_framework.database.sa_delete")
    async def test_delete(self, sa_delete: MagicMock):
        model = MagicMock()

        result = delete(model)

        sa_delete.assert_called_once_with(model)
        self.assertEqual(result, sa_delete())

    async def test_add_row(self):
        row = MagicMock()
        db: DB = DB(getenv("DB_DRIVER", "postgresql+asyncpg"), {})
        db._session.add = MagicMock()

        await db.add(row)

        db._session.add.assert_called_with(row)

    async def test_get_users(self):
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get("/users")

        self.assertIsInstance(response.json(), List)

    async def test_add_user(self):
        username = "".join(choices(ascii_letters, k=100))

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.post(f"/users/{username}")

        self.assertEqual(response.status_code, 200)
        self.assertTrue("name" in response.json())
        self.assertTrue("id" in response.json())
        self.assertEqual(response.json()["name"], username)
        self.assertIsInstance(response.json()["id"], int)

    async def test_add_user_already_exists(self):
        username = "".join(choices(ascii_letters, k=100))

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.post(f"/users/{username}")
            response2: Response = await ac.post(f"/users/{username}")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response2.status_code, 409)

    async def test_get_user_by_name(self):
        username = "".join(choices(ascii_letters, k=100))

        async with AsyncClient(app=app, base_url="https://test") as ac:
            await ac.post(f"/users/{username}")
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.get(f"/users/{username}")

        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.json(), List)

    async def test_remove_user(self):
        username = "".join(choices(ascii_letters, k=100))

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.post(f"/users/{username}")

        self.assertEqual(response.status_code, 200)

        async with AsyncClient(app=app, base_url="https://test") as ac:
            response = await ac.delete(f"/users/{username}")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), "true")

    async def test_remove_user_not_exists(self):
        async with AsyncClient(app=app, base_url="https://test") as ac:
            response: Response = await ac.delete("/users/this_username_dont_exists")

        self.assertEqual(response.status_code, 404)

    async def test_initialise_database_multiple_time(self):
        await database_dependency.init()
        await database_dependency.init()

        self.assertEqual(database_dependency.initialised, True)

    async def test_database_count(self):
        db: DB = await database_dependency()

        user_count: int = await db.count(select(User))
        users: List[Dict] = await db.all(select(User))

        self.assertEqual(len(users), user_count)

    async def test_create_tables(self):
        db: DB = await database_dependency()

        await db.create_tables()

    @patch("fastapi_framework.database.DB_POOL", False)
    async def test_init_database_dependency_without_pool(self):
        DatabaseDependency()



import logging
import sys
from unittest import TestCase
from unittest.mock import MagicMock, patch
from fastapi_framework.logger import logging_formatter, get_logger, logging_handler
from fastapi_framework import logger


class LoggerTest(TestCase):
    def test_logging_format(self):
        self.assertIsInstance(logging_formatter, logging.Formatter)
        self.assertEqual("[%(asctime)s] [%(levelname)s] %(message)s", logging_formatter._fmt)

    @patch("fastapi_framework.logger.LOG_LEVEL")
    @patch("fastapi_framework.logger.logging_handler")
    @patch("logging.getLogger")
    @patch.object(logger, "disabled_modules", [])
    def test_get_logger(self, getLogger_patch: MagicMock, logging_handler_patch: MagicMock, log_level_patch: MagicMock):
        name = MagicMock()

        result = get_logger(name)

        getLogger_patch.assert_called_with(name)
        self.assertEqual(result, getLogger_patch())
        getLogger_patch().addHandler.assert_called_with(logging_handler_patch)
        getLogger_patch().setLevel.assert_called_with(log_level_patch.upper())

    @patch.object(logger, "disabled_modules", ["logger"])
    def test_get_logger_disabled(self):
        with self.assertRaises(Exception):
            get_logger(MagicMock())

    def test_logging_handler(self):
        self.assertIsInstance(logging_handler, logging.StreamHandler)
        self.assertEqual(logging_handler.stream, sys.__stdout__)
        self.assertEqual(logging_handler.formatter, logging_formatter)



import asyncio
from typing import Set
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch, AsyncMock

from fastapi_framework.in_memory_backend import RAMBackend

ram_backend = RAMBackend()


class TestInMemoryBackend(IsolatedAsyncioTestCase):
    async def test_init_ram_backend(self):
        RAMBackend()

    async def test_set_value(self):
        await ram_backend.set("test_set_value", "test_value")

        self.assertTrue("test_set_value" in ram_backend.data)
        self.assertEqual(ram_backend.data["test_set_value"].value, b"test_value")

    async def test_get_value(self):
        await ram_backend.set("test_get_value", "test_value")

        self.assertEqual(await ram_backend.get("test_get_value"), b"test_value")

    async def test_set_value_wrong_params(self):
        with self.assertRaises(Exception):
            await ram_backend.set("test_set_value_wrong_params", "test_value", exists="WRONG")

    async def test_set_value_with_pttl(self):
        await ram_backend.set("test_set_value_with_pttl", "test_value", pexpire=2000)

        self.assertTrue("test_set_value_with_pttl" in ram_backend.data)
        pexpire = ram_backend.data["test_set_value_with_pttl"].pexpire
        self.assertIsInstance(pexpire, int)
        self.assertTrue(pexpire <= 2000)

    async def test_set_value_with_ttl(self):
        await ram_backend.set("test_set_value_with_ttl", "test_value", expire=10)

        self.assertTrue("test_set_value_with_ttl" in ram_backend.data)
        expire = int(ram_backend.data["test_set_value_with_ttl"].pexpire / 1000)
        self.assertIsInstance(expire, int)
        self.assertTrue(expire <= 10)

    async def test_get_pttl(self):
        await ram_backend.set("test_get_pttl", "test_value", expire=10)

        self.assertTrue("test_get_pttl" in ram_backend.data)

        pexpire = ram_backend.data["test_get_pttl"].pexpire

        pexpire_getter = await ram_backend.pttl("test_get_pttl")

        self.assertIsInstance(pexpire, int)
        self.assertIsInstance(pexpire, int)
        self.assertTrue(pexpire_getter <= pexpire)

    async def test_get_pttl_key_dont_exists(self):
        self.assertEqual(await ram_backend.pttl("this_key_doesn't_exists"), -2)

    async def test_get_pttl_with_expired_key(self):
        await ram_backend.set("test_get_pttl_with_expired_key", "test_value", pexpire=1)
        await asyncio.sleep(0.01)

        self.assertEqual(await ram_backend.pttl("test_get_pttl_with_expired_key"), -1)

    async def test_get_value_not_existing(self):
        self.assertEqual(await ram_backend.get("this_key_doesn't_exists"), None)

    async def test_pexpire(self):
        await ram_backend.set("test_pexpire", "test_value")
        await ram_backend.pexpire("test_pexpire", 1000)

        self.assertTrue("test_pexpire" in ram_backend.data)
        self.assertEqual(ram_backend.data["test_pexpire"].pexpire, 1000)

    async def test_pexpire_dont_exsist(self):
        self.assertEqual(await ram_backend.pexpire("this_key_doesn't_exists", 1), False)

    @patch("fastapi_framework.in_memory_backend.RAMBackend.pexpire")
    async def test_expire(self, pexpire_patch: AsyncMock):
        await ram_backend.expire("test", 15)

        pexpire_patch.assert_called_with("test", 15000)

    async def test_increase(self):
        await ram_backend.set("test_increase", 1)

        for _ in range(10):
            await ram_backend.incr("test_increase")

        self.assertEqual(await ram_backend.get("test_increase"), b"11")

    async def test_decrease(self):
        await ram_backend.set("test_decrease", 11)

        for _ in range(10):
            await ram_backend.decr("test_decrease")

        self.assertEqual(await ram_backend.get("test_decrease"), b"1")

    async def test_increase_dont_exists(self):
        for _ in range(10):
            await ram_backend.incr("test_increase_dont_exists")

        self.assertEqual(await ram_backend.get("test_increase_dont_exists"), b"10")

    async def test_decrease_dont_exists(self):
        for _ in range(10):
            await ram_backend.decr("test_decrease_dont_exists")

        self.assertEqual(await ram_backend.get("test_decrease_dont_exists"), b"-10")

    async def test_increase_with_string(self):
        await ram_backend.set("test_increase_with_string", "hello")

        with self.assertRaises(Exception):
            await ram_backend.incr("test_increase_with_string")

    async def test_decrease_with_string(self):
        await ram_backend.set("test_decrease_with_string", "hello")

        with self.assertRaises(Exception):
            await ram_backend.decr("test_decrease_with_string")

    async def test_get_expired_key(self):
        await ram_backend.set("test_get_expired_key", "test_value", pexpire=1)

        await asyncio.sleep(0.001)

        self.assertEqual(await ram_backend.get("test_get_expired_key"), None)

    async def test_get_ttl(self):
        await ram_backend.set("test_get_ttl", "test_value", expire=10)

        self.assertIsInstance(await ram_backend.ttl("test_get_ttl"), int)
        self.assertTrue(await ram_backend.ttl("test_get_ttl") <= 10)

    async def test_get_ttl_not_expiring(self):
        await ram_backend.set("test_get_ttl_not_expiring", "test_value")

        result = await ram_backend.ttl("test_get_ttl_not_expiring")
        self.assertIsInstance(result, int)
        self.assertEqual(result, -1)

    async def test_pexpire_with_expired_key(self):
        await ram_backend.set("test_pexpire_with_expired_key", "test_value", pexpire=1)
        await asyncio.sleep(0.001)

        result = await ram_backend.pexpire("test_pexpire_with_expired_key", 1)
        self.assertIsInstance(result, bool)
        self.assertEqual(result, False)

    async def test_delete(self):
        await ram_backend.set("test_delete", "test_value")

        self.assertEqual(await ram_backend.get("test_delete"), b"test_value")

        await ram_backend.delete("test_delete")

        self.assertEqual(await ram_backend.get("test_delete"), None)

    async def test_delete_dont_exists(self):
        await ram_backend.delete("test_delete_dont_exists")

    async def test_set_if_not_exists_and_set_if_exists(self):
        await ram_backend.set("test_set_if_not_exists", "1", exists=ram_backend.SET_IF_NOT_EXIST)

        self.assertEqual(await ram_backend.get("test_set_if_not_exists"), b"1")

        await ram_backend.set("test_set_if_not_exists", "2", exists=ram_backend.SET_IF_NOT_EXIST)

        self.assertEqual(await ram_backend.get("test_set_if_not_exists"), b"1")

        await ram_backend.set("test_set_if_not_exists", "3", exists=ram_backend.SET_IF_EXIST)

        self.assertEqual(await ram_backend.get("test_set_if_not_exists"), b"3")

        del ram_backend.data["test_set_if_not_exists"]
        await ram_backend.set("test_set_if_not_exists", "4", exists=ram_backend.SET_IF_EXIST)

        self.assertEqual(await ram_backend.get("test_set_if_not_exists"), None)

    async def test_smembers_with_empty_set(self):
        result: Set = await ram_backend.smembers("test_smembers_with_empty_set")

        self.assertEqual(result, set())

    async def test_sadd_item_to_set(self):
        await ram_backend.sadd("test_sadd_item_to_set", "test_value_1")

        self.assertEqual(await ram_backend.smembers("test_sadd_item_to_set"), {"test_value_1"})

    async def test_sadd_multiple_values(self):
        for i in range(5):
            await ram_backend.sadd("test_sadd_multiple_values", f"test_value_{i}")

        self.assertEqual(
            await ram_backend.smembers("test_sadd_multiple_values"), set(map(lambda x: f"test_value_{x}", range(5)))
        )

    async def test_srem(self):
        await ram_backend.sadd("test_srem", "test_srem_1")
        await ram_backend.sadd("test_srem", "test_srem_2")

        self.assertEqual(await ram_backend.smembers("test_srem"), {"test_srem_1", "test_srem_2"})

        await ram_backend.srem("test_srem", "test_srem_2")

        self.assertEqual(await ram_backend.smembers("test_srem"), {"test_srem_1"})

    async def test_srem_member_dont_exists(self):
        await ram_backend.sadd("test_srem_member_dont_exists", "test_srem_1")

        self.assertEqual(await ram_backend.srem("test_srem_member_dont_exists", "test"), False)

    async def test_srem_key_dont_exists(self):
        self.assertEqual(await ram_backend.srem("test_srem_key_dont_exists", "test"), False)

    async def test_exists(self):
        self.assertEqual(await ram_backend.exists("test_exists"), False)

        await ram_backend.set("test_exists", "test_value")

        self.assertEqual(await ram_backend.exists("test_exists"), True)

    async def test_smembers_with_string(self):
        await ram_backend.set("test_smembers_with_string", "test_value")

        self.assertEqual(await ram_backend.smembers("test_smembers_with_string"), {b"test_value"})

    async def test_increase_with_list(self):
        await ram_backend.set("test_increase_with_list", ["test"])

        with self.assertRaises(Exception):
            await ram_backend.incr("test_increase_with_list")

    async def test_decrease_with_list(self):
        await ram_backend.set("test_decrease_with_list", ["test"])

        with self.assertRaises(Exception):
            await ram_backend.decr("test_decrease_with_list")



import logging
import sys
from dotenv import load_dotenv
from os import getenv

from .modules import disabled_modules

load_dotenv()
LOG_LEVEL = getenv("LOG_LEVEL", "INFO")

logging_formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] %(message)s")

logging_handler = logging.StreamHandler(sys.stdout)
logging_handler.setFormatter(logging_formatter)


def get_logger(name: str) -> logging.Logger:
    """Get a logger with a given name."""
    if "logger" in disabled_modules:
        raise Exception("Module Logger is disabled")
    logger: logging.Logger = logging.getLogger(name)
    logger.addHandler(logging_handler)
    logger.setLevel(LOG_LEVEL.upper())

    return logger



from os import getenv
from typing import List, Set, Dict

from dotenv import load_dotenv

load_dotenv()

dependencies: Dict[str, List] = {
    "redis": [],
    "database": [],
    "logger": [],
    "jwt_auth": [],
    "rate_limit": [],
    "settings": ["database"],
    "config": [],
}
disabled_modules: List[str] = list(
    map(str.lower, getenv("DISABLED_MODULES", "").replace(" ", "").replace(",", ";").split(";"))
)


def check_dependencies():
    """Checks if Dependencies exists"""
    needed_dependencies: Set[tuple[str, str]] = set()
    for module in [dependency for dependency in dependencies if dependency not in disabled_modules]:
        for module_dependency in [
            needed_dependency for needed_dependency in dependencies[module] if needed_dependency in disabled_modules
        ]:
            needed_dependencies.add((module, module_dependency))
    if len(needed_dependencies) == 0:
        return
    raise Exception(
        f"Module '{needed_dependencies.pop()[0]}' needs the disabled Module '{needed_dependencies.pop()[1]}'"
    )



from os import getenv
from typing import Union, Optional

from sqlalchemy import Column, String

from .redis import redis_dependency, Redis
from .database import database_dependency, DB, select, Base

CACHE_TTL = int(getenv("CACHE_TTL", str(60 * 60 * 5)))


class SettingsModel(Base):
    __tablename__ = "settings"
    key: Union[str, Column] = Column(String(255), primary_key=True, unique=True)
    value: Union[str, Column] = Column(String())

    @staticmethod
    async def create(key: str, value: Union[str, int, float, bool], db: DB) -> "SettingsModel":
        row = SettingsModel(key=key, value=str(value))
        await db.add(row)
        return row


class Settings:
    """Project Settings"""

    @staticmethod
    async def set(key: str, value: Union[str, int, float, bool]):
        """Sets a Key to a Value and caches it"""
        redis: Redis = await redis_dependency()
        db: DB = await database_dependency()
        setting: SettingsModel
        if isinstance(value, bool):
            value = int(value)
        if (setting := await db.first(select(SettingsModel).filter_by(key=key))) is None:
            await SettingsModel.create(key, value, db)
        else:
            setting.value = value
        await db.commit()
        await redis.set(f"settings:{key}", value, expire=CACHE_TTL)

    @staticmethod
    async def get(key: str) -> Optional[str]:
        """Gets a Value from a Key"""
        redis: Redis = await redis_dependency()
        redis_key = f"settings:{key}"
        value: bytes
        if (value := await redis.get(redis_key)) is not None:
            return value.decode("utf-8")
        db: DB = await database_dependency()
        setting: SettingsModel = await db.first(select(SettingsModel).filter_by(key=key))
        if setting is None:
            return None
        await redis.set(redis_key, setting.value)
        return setting.value



from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, patch

from fastapi_framework.modules import check_dependencies
from fastapi_framework import modules


class TestModules(IsolatedAsyncioTestCase):
    @patch.object(modules, "disabled_modules", [])
    async def test_check_dependencies_no_disabled(self):
        check_dependencies()

    @patch.object(modules, "disabled_modules", ["database"])
    async def test_check_dependencies_redis_disabled(self):
        self.assertRaises(Exception, check_dependencies)

    @patch.object(modules, "disabled_modules", ["redis", "rate_limit", "jwt_auth"])
    async def test_check_dependencies_disabled_all_depends_on_redis(self):
        check_dependencies()



from datetime import datetime, timedelta
from os import getenv
from typing import Dict

import jwt
from .redis import Redis
from dotenv import load_dotenv
from fastapi import Depends, HTTPException
from fastapi.security import HTTPBearer
from passlib.context import CryptContext

load_dotenv()

SECRET_KEY = getenv("JWT_SECRET_KEY", "")


def check_secret_key():
    if SECRET_KEY == "":
        raise Exception("You have to set a Secret Key for JWT Auth")


check_secret_key()

ALGORITHM = getenv("JWT_ALGORITHM", "HS256")
ACCESS_TOKEN_EXPIRE_MINUTES = int(getenv("JWT_ACCESS_TOKEN_EXPIRE_MINUTES", "30"))
REFRESH_TOKEN_EXPIRE_MINUTES = int(getenv("JWT_REFRESH_TOKEN_EXPIRE_MINUTES", f"{60 * 6}"))

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

bearer_scheme = HTTPBearer()


async def get_token(token=Depends(bearer_scheme)) -> str:
    """Fastapi Dependency to get the JWT/Bearer Token"""
    return str(token.credentials)


async def create_jwt_token(data: dict, expires_delta: timedelta) -> str:
    """Creates an JWT Token with `data` and `expire_delta`"""
    data = data.copy()
    expire = datetime.utcnow() + expires_delta
    data.update({"exp": expire})
    encoded_jwt = jwt.encode(data, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


async def get_data(token: str = Depends(get_token)) -> Dict:
    """Fastapi Dependency to get JWT Data from the User"""
    try:
        data = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
    except jwt.exceptions.InvalidTokenError as e:
        if isinstance(e, jwt.exceptions.ExpiredSignatureError):
            raise HTTPException(status_code=401, detail="Token is expired")
        raise HTTPException(status_code=401, detail="Token is invalid")
    return data


async def create_access_token(data: Dict) -> str:
    """Creates an JWT Access Token with `data` and expires after `JWT_ACCESS_TOKEN_EXPIRE_MINUTES`"""
    return await create_jwt_token(data, timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES))


async def create_refresh_token(user_id: int, redis: Redis) -> str:
    """Creates an Refresh Token with the `user_id` and needs `redis`"""
    refresh_token_data: Dict = {"user_id": user_id}
    refresh_token: str = await create_jwt_token(refresh_token_data, timedelta(minutes=REFRESH_TOKEN_EXPIRE_MINUTES))
    await redis.sadd("refresh_tokens", refresh_token)
    return refresh_token


async def invalidate_refresh_token(refresh_token: str, redis: Redis) -> None:
    """Invalidates a Refresh Token"""
    await redis.srem("refresh_tokens", refresh_token)


async def check_refresh_token(refresh_token: str, redis: Redis) -> bool:
    """Checks if a Refresh Token is valid (in Redis Cache)"""
    refresh_tokens = [refresh_token.decode("utf-8") for refresh_token in await redis.smembers("refresh_tokens")]
    if refresh_token in refresh_tokens:
        return True
    return False


async def generate_tokens(data: Dict, user_id: int, redis: Redis) -> Dict:
    """Generates Access and Refresh Token with `data`, `user_id` and needs redis"""
    access_token: str = await create_access_token(data)
    refresh_token: str = await create_refresh_token(int(user_id), redis)
    return {"access_token": access_token, "refresh_token": refresh_token, "token_type": "bearer"}



from setuptools import setup, find_packages

with open("requirements.txt", "r") as file:
    requirements = [line.strip() for line in file.readlines()]

with open("README.md", "r") as file:
    readme = file.read().replace("[ ]", "❌").replace("[x]", "✅")

setup(
    name="fastapi-framework",
    packages=find_packages(),
    version="1.0.0",
    description="A Fastapi Framework",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Tert0",
    license="MIT",
    install_requires=requirements,
    extra_requires=["mkdocs-material", "black", "coverage"],
    url="https://github.com/Tert0/fastapi-framework",
    python_requires=">=3.6",
)



from os import getenv
from typing import TypeVar, Dict

from dotenv import load_dotenv
from sqlalchemy.engine import URL
from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine, AsyncSession
from sqlalchemy.future import select as sa_select
from sqlalchemy.orm import DeclarativeMeta, declarative_base, sessionmaker
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import Executable
from sqlalchemy.sql.expression import exists as sa_exists, delete as sa_delete, Delete
from sqlalchemy.sql.functions import count
from sqlalchemy.sql.selectable import Select, Exists

from .logger import get_logger

load_dotenv()

T = TypeVar("T")


def select(entity) -> Select:
    """Shortcut for :meth:`sqlalchemy.future.select`"""

    return sa_select(entity)


def filter_by(cls, *args, **kwargs) -> Select:
    """Shortcut for :meth:`sqlalchemy.future.Select.filter_by`"""
    return select(cls, *args).filter_by(**kwargs)


def exists(*entities, **kwargs) -> Exists:
    """Shortcut for :meth:`sqlalchemy.sql.expression.exists`"""
    return sa_exists(*entities, **kwargs)


def delete(table) -> Delete:
    """Shortcut for :meth:`sqlalchemy.sql.expression.delete`"""
    return sa_delete(table)


class DB:
    """An async SQLAlchemy ORM wrapper"""

    Base: DeclarativeMeta
    _engine: AsyncEngine
    _session: AsyncSession

    def __init__(self, driver: str, options: Dict = {"pool_size": 20, "max_overflow": 20}, **kwargs):
        url: str = URL.create(drivername=driver, **kwargs)
        self._engine = create_async_engine(url, echo=True, pool_pre_ping=True, pool_recycle=300, **options)
        self.Base = declarative_base()
        self._session: AsyncSession = sessionmaker(self._engine, expire_on_commit=False, class_=AsyncSession)()

    async def create_tables(self):
        """Creates all Model Tables"""
        async with self._engine.begin() as conn:
            await conn.run_sync(self.Base.metadata.create_all)

    async def add(self, obj: T) -> T:
        """Adds an Row to the Database"""
        self._session.add(obj)
        return obj

    async def delete(self, obj: T) -> T:
        """Deletes a Row from the Database"""
        await self._session.delete(obj)
        return obj

    async def exec(self, statement: Executable, *args, **kwargs):
        """Executes a SQL Statement"""
        return await self._session.execute(statement, *args, **kwargs)

    async def stream(self, statement: Executable, *args, **kwargs):
        """Returns an Stream of Query Results"""
        return (await self._session.stream(statement, *args, **kwargs)).scalars()

    async def all(self, statement: Executable, *args, **kwargs) -> list[T]:
        """Returns all matches for a Query"""
        return [x async for x in await self.stream(statement, *args, **kwargs)]

    async def first(self, *args, **kwargs):
        """Returns first match for a Query"""
        return (await self.exec(*args, **kwargs)).scalar()

    async def exists(self, *args, **kwargs):
        """Checks if there is a match for this Query"""
        return await self.first(exists(*args, **kwargs).select())

    async def count(self, *args, **kwargs):
        """Counts matches for a Query"""
        return await self.first(select(count()).select_from(*args, **kwargs))

    async def commit(self):
        """Commits/Saves changes to Database"""
        await self._session.commit()


logger = get_logger(__name__)

DB_HOST = getenv("DB_HOST", "localhost")
DB_PORT = getenv("DB_PORT", "5432")
DB_DATABASE = getenv("DB_DATABASE")
DB_USERNAME = getenv("DB_USERNAME", "postgres")
DB_PASSWORD = getenv("DB_PASSWORD", "")
DB_POOL_SIZE = getenv("DB_POOL_SIZE", "20")
DB_MAX_OVERFLOW = getenv("DB_MAX_OVERFLOW", "20")
DB_POOL = True if getenv("DB_POOL", "True").lower() == "true" else False


class DatabaseDependency:
    db: DB
    database_url_options: Dict
    engine_options: Dict
    initialised: bool = False

    def __init__(self):
        self.database_url_options = {
            "host": DB_HOST,
            "port": DB_PORT,
            "database": DB_DATABASE,
            "username": DB_USERNAME,
            "password": DB_PASSWORD,
        }
        self.database_url_options = dict([(k, v) for k, v in self.database_url_options.items() if v != ""])
        self.engine_options: Dict = {
            "pool_size": DB_POOL_SIZE,
            "max_overflow": DB_MAX_OVERFLOW,
            "poolclass": DB_POOL,
        }
        self.engine_options = dict([(k, int(v)) for k, v in self.engine_options.items() if v != ""])
        if self.engine_options["poolclass"] == 0:
            self.engine_options["poolclass"] = NullPool
        else:
            del self.engine_options["poolclass"]
        self.db = DB(
            getenv("DB_DRIVER", "postgresql+asyncpg"), options=self.engine_options, **self.database_url_options
        )
        logger.info("Connected to Database")

    async def init(self) -> None:
        if self.initialised:
            return
        logger.info("Create Tables")
        self.initialised = True
        await self.db.create_tables()

    async def __call__(self) -> DB:
        if not self.initialised:
            await self.init()
        return self.db


database_dependency: DatabaseDependency = DatabaseDependency()
Base: DeclarativeMeta = database_dependency.db.Base



"""A FastAPI Framework for things like Database, Redis, Logging, JWT Authentication and Rate Limits"""

__version__ = "1.4.6.1"

from .modules import check_dependencies, disabled_modules

check_dependencies()  # noqa: FLK-E402
from .database import database_dependency
from .jwt_auth import (
    create_jwt_token,
    create_access_token,
    create_refresh_token,
    invalidate_refresh_token,
    get_token,
    get_data,
    pwd_context,
    ACCESS_TOKEN_EXPIRE_MINUTES,
    REFRESH_TOKEN_EXPIRE_MINUTES,
    check_refresh_token,
    generate_tokens,
)
from .logger import get_logger
from .rate_limit import RateLimitManager, RateLimiter, get_uuid_user_id, RateLimitTime
from .redis import get_redis, RedisDependency, redis_dependency, Redis
from .in_memory_backend import InMemoryBackend, RAMBackend
from .config import Config, ConfigField
